/* $Workfile: OsaString.cpp $   $Revision: 8 $    $Modtime: 8/09/04 5:42p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 1998-2003, All Rights Reserved

*******************************************************************************/

#include "OsaRtl.h"
#include "OsaString.h"
#include "OsaMemory.h"

/******************************************************************************
******************************************************************************/
extern "C" int OsaStrLen(const char *s)
{
  int i;

  for (i = 0; s[i] != '\0'; i++)
    ;
  return i;
}

/******************************************************************************
******************************************************************************/
extern "C" void OsaStrCpy(char *s, char *t)
{
  for (; (*s = *t) != '\0'; s++, t++)
    ;
}

/******************************************************************************
******************************************************************************/
extern "C" int OsaStrCmp(const char *s, const char *t)
{
  for (; *s == *t; s++, t++) {
    if (*s == '\0') {
      return 0;
    }
  }

  return *s - *t;
}

/******************************************************************************
******************************************************************************/
extern "C" OsaUint32 OsaWcsToUint32(const wchar_t *pwcs, const wchar_t **ppwcEnd, int nBase)
{
  OsaUint32 num = 0;
  const wchar_t *pwc = pwcs;

  while (*pwc == L' ' || *pwc == L'\t') ++pwc;

  if (*pwc != L'\0') {
    // Decimal
    switch (nBase) {
    case 10:
      while (L'0' <= *pwc && *pwc <= L'9') {
        num = (num * 10) + (*pwc++ - L'0');
      }
      break;
    case 16:
      for (; L'0' <= *pwc && *pwc <= L'9' || L'A' <= towupper(*pwc) && towupper(*pwc) <= L'F'; ++pwc) {
        if (L'0' <= *pwc && *pwc <= L'9') {
          num = (num * 16) + (*pwc - L'0');
        } else {
          num = (num * 16) + (towupper(*pwc) - L'A' + 10);
        }
      }
      break;
    case 2:
      for (pwc; *pwc == L'0' || *pwc == L'1'; ++pwc) {
        num = (num * 2) + (*pwc - L'0');
      }
      break;
    case 8:
      while (L'0' <= *pwc && *pwc <= L'7') {
        num *= 8;
        num += *pwc++ - L'0';
      }
      break;
    default:
      if (ppwcEnd) {
        *ppwcEnd = pwcs;
      }
      break;
    }
  }

  if (ppwcEnd) {
    *ppwcEnd = pwc;
  }

  return num;
}


/******************************************************************************
  OsaString functions
******************************************************************************/


OsaString &OsaString::Append(unsigned int nValue)
{
  const int BUF_SIZE = 32;
  wchar_t buffer[BUF_SIZE];
  buffer[BUF_SIZE - 1] = L'\0';
  int i = BUF_SIZE - 2;
  do {
    buffer[i] = L'0' + (nValue % 10);
    nValue = nValue / 10;
    i--;
  } while (nValue);

  return Append(buffer + i + 1, (BUF_SIZE - 2) - i);  // Using Append(s,c) with length instead of  Append(s) so we don't have to calculate the length twice
}


/******************************************************************************
  OsaStringRep functions
******************************************************************************/

OsaStringRep::OsaStringRep(const char *pszString)
{
	m_cbUsed = OsaStrLen(pszString) * sizeof(wchar_t);

  m_cbBufSize    = 2 * m_cbUsed;
  m_pwcszBuffer = new wchar_t[(m_cbBufSize / sizeof(wchar_t)) + 1];
  for (unsigned int iChar = 0; iChar < (m_cbUsed/sizeof(wchar_t) + 1); iChar++) {
    m_pwcszBuffer[iChar] = (wchar_t) pszString[iChar];
  }
}


OsaStringRep::OsaStringRep(const wchar_t *pwcsString, unsigned int coCharsInString)
{
  m_cbUsed = static_cast<unsigned short>(coCharsInString * sizeof(wchar_t));
  m_cbBufSize    = 2 * m_cbUsed;
  m_pwcszBuffer = new wchar_t[(m_cbBufSize / sizeof(wchar_t)) + 1];
  wcsncpy(m_pwcszBuffer, pwcsString, m_cbUsed / sizeof(wchar_t));
  m_pwcszBuffer[m_cbUsed / sizeof(wchar_t)] = L'\0';
}


OsaStringRep &OsaStringRep::operator=(const OsaStringRep &that)
{
  if (this != &that) {
    // Reallocate if size difference is too great
    if ((m_cbBufSize < 2 * that.m_cbUsed) || m_cbBufSize > 4 * that.m_cbUsed) {
      delete m_pwcszBuffer;
      m_cbBufSize = 2 * that.m_cbUsed;
      m_pwcszBuffer = new wchar_t[(m_cbBufSize / sizeof(wchar_t)) + 1];
    }
    ::wcscpy(m_pwcszBuffer, that.m_pwcszBuffer);
    m_cbUsed = that.m_cbUsed;
  } 
  return *this;
}


bool OsaStringRep::Reallocate(unsigned int coNewBufSize)
{
  // Reallocate to new size, truncate original contents if necessary
  wchar_t *pwcszBuffer = m_pwcszBuffer;
  m_cbBufSize = static_cast<unsigned short>(coNewBufSize * sizeof(wchar_t));
  if (m_pwcszBuffer = new wchar_t[(m_cbBufSize / sizeof(wchar_t)) + 1]) {
    if (m_cbUsed < m_cbBufSize) {
      ::wcscpy(m_pwcszBuffer, pwcszBuffer);
    } else {
      // Truncate
      ::wcsncpy(m_pwcszBuffer, pwcszBuffer, (m_cbBufSize / sizeof(wchar_t)) - 1);
      m_cbUsed = m_cbBufSize - sizeof(wchar_t);
      m_pwcszBuffer[m_cbUsed / sizeof(wchar_t)] = L'\0';
    }
    delete pwcszBuffer;
    return true;
  } else {
    return false;
  }
}

OsaStringRep &OsaStringRep::Append(const char *pszString)
{
	size_t coNewChars = OsaStrLen(pszString);

  if (EnsureEnoughRoomForAdding(coNewChars)) {
    for (unsigned int iChar = 0; iChar < coNewChars + 1; iChar++) {
      m_pwcszBuffer[(m_cbUsed / sizeof(wchar_t)) + iChar] = (wchar_t) pszString[iChar];
    }
    m_cbUsed += static_cast<unsigned short>(coNewChars * sizeof(wchar_t));
  }
  return *this;
}


OsaStringRep &OsaStringRep::Append(const wchar_t *pwcsString, unsigned int coCharsInString)
{
  if (m_pwcszBuffer == pwcsString) {
    // Force early possible realloc and make the pointers equal again
    // Note: A realloc would make use of new memory and suddenly
    // pwcsString would point into free'd memory
    EnsureEnoughRoomForAdding(coCharsInString);
    pwcsString = m_pwcszBuffer;
  }
  if (EnsureEnoughRoomForAdding(coCharsInString)) {
    ::wcsncpy(m_pwcszBuffer + (m_cbUsed / sizeof(wchar_t)), pwcsString, coCharsInString);
    m_cbUsed += static_cast<unsigned short>(coCharsInString * sizeof(wchar_t));
    m_pwcszBuffer[m_cbUsed / sizeof(wchar_t)] = L'\0';
  }
  return *this;
}



#ifdef OSA_TEST

int main()
{
  OsaString str;

  str = "Hello";
  str.Append(" World");
  str.Append(L"!");

  str.Append(str);

  str = "OtxDevice";
  str.Append(3);

  return 0;
}

#endif

